#include <ArduinoBLE.h>
#include <SD.h>

// Names for Pins
#define releaseButton 2
#define secureButton 3
#define cameraButton 4
#define buzzer 5
#define solenoidPin1 8
#define solenoidPin2 9
#define chipSelect 10
#define irReciever1 A5
#define irReciever2 A6
#define irReciever3 A7

#define DEBOUNCE_DELAY 50
#define IR_DELAY 1
unsigned long lastReleaseButtonDebounceTime = 0;
unsigned long lastSecureButtonDebounceTime = 0;
unsigned long lastMasterIrSensorDebounceTime = 0;

int releaseButtonState = HIGH;
int lastReleaseButtonState = LOW;
int secureButtonState = HIGH;
int lastSecureButtonState = LOW;
int masterIrSensorState = HIGH;
int lastMasterIrSensorState = LOW;

int numberSnapshots;
#define maxTrials 25
int trialNumber = 1;
bool trialStart = false;
char receivedBuffer[255];
String imuData[202];

unsigned long startTimer;
unsigned long endTimer;
unsigned long totalTime;




void conductTrials(BLEDevice peripheral) {
  // connect to the peripheral
  Serial.println("Connecting ...");

  if (peripheral.connect()) {
    Serial.println("Connected");
  } else {
    Serial.println("Failed to connect!");
    return;
  }

  // discover peripheral attributes
  Serial.println("Discovering attributes ...");
  if (peripheral.discoverAttributes()) {
    Serial.println("Attributes discovered");
  } else {
    Serial.println("Attribute discovery failed!");
    peripheral.disconnect();
    return;
  }

  // retrieve the IMU characteristics
  BLECharacteristic activateIMUCharacteristic = peripheral.characteristic("0001");
  BLECharacteristic transferIMUDataCharacteristic = peripheral.characteristic("0002");

  if (!activateIMUCharacteristic) {
    Serial.println("Peripheral does not have an Activate IMU characteristic!");
    peripheral.disconnect();
    return;
  } else if (!activateIMUCharacteristic.canWrite()) {
    Serial.println("Peripheral does not have a writable Activate IMU characteristic!");
    peripheral.disconnect();
    return;
  }

  // Subscribe to the Transfer IMU Data characteristic
  Serial.println("Subscribing to Transfer IMU Data characteristic ...");
  if (!transferIMUDataCharacteristic) {
    Serial.println("no Transfer IMU Data characteristic found!");
    peripheral.disconnect();
    return;
  } else if (!transferIMUDataCharacteristic.canSubscribe()) {
    Serial.println("Transfer IMU Data characteristic is not subscribable!");
    peripheral.disconnect();
    return;
  } else if (!transferIMUDataCharacteristic.subscribe()) {
    Serial.println("subscription failed!");
    peripheral.disconnect();
    return;
  } else {
    Serial.println("Subscribed");
  }

  while (peripheral.connected()) {
    if (trialNumber <= maxTrials) {

      int releaseButtonReading = digitalRead(releaseButton);
      int secureButtonReading = digitalRead(secureButton);
      int irSensor1Reading = digitalRead(irReciever1);
      int irSensor2Reading = digitalRead(irReciever2);
      int irSensor3Reading = digitalRead(irReciever3);
      int masterIrSensorReading = HIGH;

      if ((irSensor1Reading == LOW) || (irSensor2Reading == LOW) || (irSensor3Reading == LOW)) {   
        masterIrSensorReading = LOW;
      }

      if (releaseButtonReading != lastReleaseButtonState) {
        // reset the debouncing timer
        lastReleaseButtonDebounceTime = millis();
      }
      if (secureButtonReading != lastSecureButtonState) {
        // reset the debouncing timer
        lastSecureButtonDebounceTime = millis();
      }
      if (masterIrSensorReading != lastMasterIrSensorState) {
        lastMasterIrSensorDebounceTime = millis();
      }

      if ((millis() - lastReleaseButtonDebounceTime) > DEBOUNCE_DELAY) {
        // whatever the reading is at, it's been there for longer than the debounce
        // delay, so take it as the actual current state:

        // if the button state has changed:
        if (releaseButtonReading != releaseButtonState) {
          releaseButtonState = releaseButtonReading;

          // only toggle the LED if the new button state is HIGH
          if (releaseButtonState == HIGH) {
            releaseObject(activateIMUCharacteristic);
            trialStart = true;
            Serial.println("Object Released");
          }
        }
      }
      if ((millis() - lastSecureButtonDebounceTime) > DEBOUNCE_DELAY) {
        // whatever the reading is at, it's been there for longer than the debounce
        // delay, so take it as the actual current state:

        // if the button state has changed:
        if (secureButtonReading != secureButtonState) {
          secureButtonState = secureButtonReading;

          // only toggle the LED if the new button state is HIGH
          if (secureButtonState == HIGH) {
            secureObject();
            Serial.println("Object Secured");
          }
        }
      }
      if ((millis() - lastMasterIrSensorDebounceTime) > IR_DELAY) {
        // whatever the reading is at, it's been there for longer than the debounce
        // delay, so take it as the actual current state:

        // if the button state has changed:
        if (masterIrSensorReading != masterIrSensorState) {
          masterIrSensorState = masterIrSensorReading;

          // only toggle the LED if the new button state is HIGH
          if ((masterIrSensorState == LOW) && (trialStart == true)) {
            endTrial();
            Serial.println("Trial Ended");
            // Activate Buzzer to Signal Trial Successful
            tone(buzzer, 4500, 1250);
            delay(1250);
            noTone(buzzer);
            recieveImuData(transferIMUDataCharacteristic);
            Serial.println("IMU Data Transmitted");
            writeCsvFile();
            trialStart = false;
            trialNumber++;
          }
        }
      }

      lastReleaseButtonState = releaseButtonReading;
      lastSecureButtonState = secureButtonReading;
      lastMasterIrSensorState = masterIrSensorReading;

    } else {
      // Activate Buzzer to Signal Series of Trials Complete
      tone(buzzer, 4500, 5000);
      delay(5000);
      noTone(buzzer);
    }
  }
}


void releaseObject(BLECharacteristic activateIMUCharacteristic) {
  // Turn on Camera
  digitalWrite(cameraButton, LOW);
  delay(10);
  digitalWrite(cameraButton, HIGH);
  delay(2500);
  // Tell IMU to Start Recording
  activateIMUCharacteristic.writeValue((byte)0x01);
  // Start Timer
  startTimer = millis();
  // Tell Solenoid to Release Object
  digitalWrite(solenoidPin1, LOW);
  digitalWrite(solenoidPin2, LOW);
}
void secureObject() {
  // Action for securing the object
  digitalWrite(solenoidPin1, HIGH);
  digitalWrite(solenoidPin2, LOW);
}
void endTrial() {
  // Stop Timer
  endTimer = millis();
  totalTime = endTimer - startTimer;
  // Turn off Camera
  delay(100);
  digitalWrite(cameraButton, LOW);
  delay(10);
  digitalWrite(cameraButton, HIGH);
}
void recieveImuData(BLECharacteristic transferIMUDataCharacteristic) {
  for (int i = 0; i < 203; i++) {
    while (!transferIMUDataCharacteristic.valueUpdated());
    memset(receivedBuffer, 0, sizeof(receivedBuffer));
    transferIMUDataCharacteristic.readValue(receivedBuffer, sizeof(receivedBuffer));
    if (i < 202) {
      imuData[i] = String(receivedBuffer); // Convert to Arduino String
      Serial.println(imuData[i]);
    } else {
      // Calculate Number of IMU Snapshots Plus Header
      numberSnapshots = (totalTime / String(receivedBuffer).toInt()) + 2;
    }
  }
}
void writeCsvFile() {
  String filename = "Trial-" + String(trialNumber) + ".csv";
  File file = SD.open(filename, FILE_WRITE);

  // if the file opened okay, write to it:
  if (file) {
    file.println("Time in Milliseconds");
    file.println(totalTime);
    Serial.print("Writing to ");
    Serial.print(filename);
    Serial.print("...");
    for (int i = 0; i < numberSnapshots; i++) {
    //for (int i = 0; i < 202; i++) {
      file.println(imuData[i]);
    }
    // close the file:
    file.close();
    Serial.println("done.");
  } else {
    // if the file didn't open, print an error:
    Serial.print("Error opening ");
    Serial.println(filename);
  }
}


void setup() {
  Serial.begin(115200);
  delay(5000);
  //while (!Serial);

  if (!BLE.begin()) {
    Serial.println("Bluetooth initialization failed!");
    while (1);
  }

  if (!SD.begin(10)) {
    Serial.println("microSD initialization failed!");
    // Activate Buzzer to Signal microSD Card Reader Failure
    tone(buzzer, 4500, 1250);
    delay(1250);
    noTone(buzzer);
    while (1);
  }

  pinMode(releaseButton, INPUT_PULLUP);
  pinMode(secureButton, INPUT_PULLUP);
  pinMode(cameraButton, OUTPUT);
  pinMode(buzzer, OUTPUT);
  pinMode(irReciever1, INPUT_PULLUP);
  pinMode(irReciever2, INPUT_PULLUP);
  pinMode(irReciever3, INPUT_PULLUP);

  // Start Scanning for Peripherals
  BLE.scanForUuid("19B10000-E8F2-537E-4F6C-D104768A1214");
}

void loop() {
  // Check if a Peripheral has been Discovered
  BLEDevice peripheral = BLE.available();

  if (peripheral) {
    // Discovered a Peripheral, Print Out Address, Local Name, and Advertised Service
    Serial.print("Found ");
    Serial.print(peripheral.address());
    Serial.print(" '");
    Serial.print(peripheral.localName());
    Serial.print("' ");
    Serial.print(peripheral.advertisedServiceUuid());
    Serial.println();

    if (peripheral.localName() != "Free-Fall Object") {
      return;
    }

    // Stop Scanning
    BLE.stopScan();

    conductTrials(peripheral);

    // Peripheral Disconnected, Start Scanning Again
    BLE.scanForUuid("19B10000-E8F2-537E-4F6C-D104768A1214");
  }
}
